
/**
 * Burton Rosenberg
 * July 2004
 */

import java.util.* ; // HashMap
import java.io.* ;

/*
   The three responsibilities of this class are:

   (1) mapping "rawtext" to "plaintext" (cipher and plain texts are equal)
       A HashMap is used.

   (2) mapping plaintext to an integer code and inversely
       A HashMap and an array are used.

   (3) be a stream of char source pure for playfair, i.e. "x" between
       double characters if even-aligned, final padding if stream odd
       length. Wraps around Reader System.in by default; around a
       String (perhaps File) by appropriate Construction.

 */


public class PlayfairInputStream {

    // rawCharacters[i] <=> plainCharacters[i]
    // code for plainCharacter[i] assigned sequentially from
    // 1 in order of distinct charactesr appearing in 
    // plainCharacters

    final static char [] rawCharacters = {
         'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j',
         'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't',
         'u', 'v', 'w', 'x', 'y', 'z',
         'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J',
         'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T',
         'U', 'V', 'W', 'X', 'Y', 'Z',
        } ;

    final static char [] plainCharacters = {
         'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'i',
         'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't',
         'u', 'v', 'w', 'x', 'y', 'z',
         'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'i',
         'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't',
         'u', 'v', 'w', 'x', 'y', 'z',
        } ;

   static final char CHAR_X = 'x' ;
   static final char CHAR_Y = 'y' ;
   static final int countPlainCharacters = 25 ;

   HashMap rawtextToPlaintext ;
   HashMap plaintextToCode ;
   char [] plaintextTable ;
   BufferedReader br ;

   /*
      Raw -> Plaintext map, hashed in forward direction
      Plaintext (default range of Raw->Plaintext) encoding (default
         canonical) hashed forward and back
   */


  void makeHashTables() {

     rawtextToPlaintext = new HashMap( rawCharacters.length * 2 ) ;    
     plaintextToCode = new HashMap( plainCharacters.length * 2 ) ;    
     plaintextTable = new char[countPlainCharacters] ;
     int code = 0 ;
     for ( int i=0; i<rawCharacters.length; i++ ) {
        Character pc = new Character(plainCharacters[i]) ;
        rawtextToPlaintext.put( new Character(rawCharacters[i]), pc ) ;
        if ( plaintextToCode.get(pc)==null ) {
           plaintextToCode.put( pc, new Integer(code) ) ;
           plaintextTable[code] = plainCharacters[i] ;
           code++ ;
        }
     }
  }

   public PlayfairInputStream() {
     makeHashTables() ;
     br = new BufferedReader( new InputStreamReader(System.in) ) ;
  }

   public PlayfairInputStream(String charSource) {
     makeHashTables() ;
     br = new BufferedReader( new StringReader(charSource) ) ;
  }

  char convertRawtextToPlaintext( char c ) {
     Character co = (Character) rawtextToPlaintext.get(new Character(c)) ;
     if ( co==null ) return (char) 0 ;
     return co.charValue() ;
  }

  public int plaintextEncode(char c) {
     Integer i = (Integer) plaintextToCode.get(new Character(c)) ;
     if ( i==null ) return -1 ;
     return i.intValue() ;
  }

  public char plaintextDecode(int i) {
     char c ;
     try {
        c = plaintextTable[i] ;
     }
     catch ( ArrayIndexOutOfBoundsException aioobe )
        { c = (char) 0 ; }
     return c ;
  }

  public int convertRawtextToPlaintext( StringBuffer sb ) {
     int j = 0 ;
     for ( int i=0; i<sb.length(); i++ ) {
        char c = convertRawtextToPlaintext(sb.charAt(i)) ;
        if ( c!= ((char) 0) ) sb.setCharAt(j++, c );
     }
     sb.setLength(j) ;
     return j ;
  }

  
  char c_remembered ;
  char c_defered ;
  boolean evenChar = true ;
  boolean c_deferedFlag = false ;

  int getCharAux() {
     int c_int ;
     char c ;
     try { while ( true ) {
        c_int = br.read() ;
        if ( c_int==-1 ) return c_int ;
        c = convertRawtextToPlaintext( (char) c_int ) ;
        if ( c!=(char)  0 ) return c ;
     } } catch ( IOException ioe ) { } 
     return -1 ;
  }

  char getNextChar() {
     char c_ret = (char) 0 ;
     int c_int ;

        if ( evenChar ) {
            if ( c_deferedFlag ) {
               c_int = c_defered ;
               c_deferedFlag = false ;
            }
            else c_int = getCharAux() ;
            c_ret = ( c_int==-1 ) ? (char) 0 : (char) c_int ;
            c_remembered = c_ret ;
            evenChar = false ;
        }
        else { // oddChar
           evenChar = true ;
           c_int = getCharAux() ;
           if ( c_int==-1 ) {
               c_ret = ( c_remembered==CHAR_X ) ? CHAR_Y : CHAR_X ; 
           }
           else if ( ((char)c_int)==c_remembered ) {
               c_ret = ( c_remembered==CHAR_X ) ? CHAR_Y : CHAR_X ; 
               c_defered = (char) c_int ;
               c_deferedFlag = true ;
           }
           else c_ret = (char) c_int ;
        }
        // fi evenChar

     return c_ret ;
  }


  // test routine, stdin to stdout, filters raw to plaintext
  public static void main(String [] args) {
      char c ;
      PlayfairInputStream pis = new PlayfairInputStream( "AbJJIm .,sTuv") ;
      while ((c=pis.getNextChar())!=(char) 0) 
          System.out.println(c) ;
      pis = new PlayfairInputStream() ;
      while ((c=pis.getNextChar())!=(char) 0) 
          System.out.println(c) ;
  
  }

}

